#!/usr/bin/env python
#  -*- coding: utf-8 -*-

""" Loaders classes

A :class:`Loader` class should be instantiated with a connection string,
to enable connection to the storage the data should be loaded to.

Once instantiated, invoking the :func:`load()` method will open the resource,
load all of the data found into the ``self.data`` dictionary and return a
dictionary of results back to the caller.

The ``self.data`` dictionary allows the :func:`load()` method to prepare
data for loading.

Details of the pre-loading operations foe each single subset of data should
be delegated to methods in the ETL object.
"""
import json
from abc import ABCMeta, abstractmethod
from collections import Iterable
import logging
import logging.config
import os
import resource
import time

try:
    from elasticsearch import Elasticsearch
    from elasticsearch.helpers import streaming_bulk
except ImportError as error:
    print(f"Warning - {error.__class__.__name__}: {error.name}." )

import pandas as pd

__author__ = "guglielmo"

from pandas import DataFrame


class Loader(object):
    """Base abstract class. All Loaders classes extend from this.
    """

    __metaclass__ = ABCMeta

    etl = None
    logger = logging.getLogger(__name__)

    @abstractmethod
    def load(self, **kwargs):
        """Base extract method, must be instantiated in extended
        classes.
        """
        pass


class CSVLoader(Loader):
    """:class:`Loader` that stores data in a CSV file

    The csv filename is specified in the constructor::

        csvloader = CSVLoader('/tmp/test.csv')

    """

    def __init__(self, csv_path, label="dati", encoding="utf8", sep=";"):
        """Create a new instance of the loader

        Needs a csv path and may accept an encoding,
        if different from UTF8.

        Args:
            csv_path (string): The path where csv files will be stored
            label: (Optional[string]): A label used to name the file
              This is only used when a single DataFrame is being loaded
            encoding: (Optional[string]): Encoding used while writing the CSV
        """

        super(CSVLoader, self).__init__()
        self.csv_path = csv_path
        self.label = label
        self.encoding = encoding
        self.sep = sep

    def load(self, **kwargs):
        """Writes data to csv file.

        The data contained in the `processed_data`` attribute of
        the `etl` instance are written in one or more csv files, in the
        `csv_path` specified when instantiating the loader.

        All `DataFrame` in the `etl.processed_data` attribute are packed and
        written into CSV.

        If `etl.processed_data` is a dictionary, then multiple files are
        written, and they're named according to `etl.processed_data`'s keys,
        with a `.csv` extension added.

        If `etl.processed_data` is a `DataFrame`, then the `label` attribute
        of the loader is used as the only label, and the name of the csv
        file will be created out of that label.

        """

        # build data dictionary
        data = {}
        if not isinstance(self.etl.processed_data, pd.DataFrame):
            if isinstance(self.etl.processed_data, dict):
                for label, df in self.etl.processed_data.items():
                    data[label] = df
            elif (
                isinstance(self.etl.processed_data, list) and
                isinstance(self.etl.processed_data[0], dict)
            ):
                data[self.label] = pd.DataFrame(self.etl.processed_data)
            else:
                raise Exception(
                    "Could not build dataframe out of processed_data of type {0}".format(type(self.etl.processed_data))
                )
        else:
            data[self.label] = self.etl.processed_data

        # create path to store csv if not existing
        try:
            os.stat(self.csv_path)
        except FileNotFoundError:
            os.mkdir(self.csv_path)

        # write data to CSV
        for label, df in data.items():
            csv_filename = "{0}/{1}.csv".format(self.csv_path, label)
            df.to_csv(csv_filename, encoding=self.encoding, sep=self.sep, index=False)


class JsonLoader(Loader):
    """:class:`Loader` that stores data in a local JSON file

    The json filename and location are specified in the constructor::

        loader = JSONLoader(out_path='/tmp', label="test")

    """

    def __init__(self, filepath: str, encoding: str = "utf8"):
        """Create a new instance of the loader

        Needs a csv path and may accept an encoding,
        if different from UTF8.

        Args:
            filepath: The complete file with path where json file will be stored
        """

        super(JsonLoader, self).__init__()
        self.filepath = filepath
        self.encoding = encoding

    def load(self, **kwargs):
        """Writes data to json file.

        The data contained in the `processed_data`` attribute of
        the `etl` instance are written in one json file, in the
        `out_path` specified when instantiating the loader.

        kwargs: arguments passed to the json.dump function
                see: https://docs.python.org/3.5/library/json.html#basic-usage
        """
        try:
            with open(self.filepath, "w", encoding=self.encoding, **kwargs) as json_file:
                if isinstance(self.etl.processed_data, DataFrame):
                    self.etl.processed_data.to_json(json_file, orient='records')
                else:
                    json.dump(self.etl.processed_data, json_file)
        except Exception as e:
            self.logger.error("Data could not be exported as json: {0}.".format(e))


class ESLoader(Loader):
    """:class:`Loader` that stores data of a given type into an Elasticsearch
    instance

    The elasticsearch constructor needs three parameters::

        esloader = ESLoader(
            'http://user:pass@localhost:9200/',
            'politici', 'person')

    """

    def __init__(
        self,
        es_hosts,
        es_index,
        es_doc_type,
        es_id_field=None,
        es_mapping=None,
        es_batchsize=500,
        es_delete=False,
        log_level="info",
    ):
        """Create a new instance of the loader

        Needs a
        connection_url, the index and document type names,
        also accepts some optional arguments

        Args:
            es_url (string): The connection url (eg
                http://username@password:localhost:9200/)
            es_index: (string): Index name
            es_doc_type: (string): Document type
            es_id_field: (Optional[string]): Name of the unique id field.
                This is optional, but if not specified, an automatic ID will
                be used and previously inserted documents will not be recognized.
            es_mapping: (Optional[dict]): a mapping that allows you to
                specify how each field will be indexed. If not specified,
                the dynamic mapping takes place.
            es_batchsize: (Optional[integer]): Number of document to be
                indexed in a single bulk call. Defaults to 500.
            es_delete: (Optional[bool]): If the index needs to be deleted
                before bulk operations.
        """

        super(ESLoader, self).__init__()
        self.log_level = log_level
        self.es_hosts = es_hosts
        self.es_index = es_index
        self.es_doc_type = es_doc_type
        self.es_id_field = es_id_field
        self.es_mapping = es_mapping
        self.es_batchsize = es_batchsize
        self.es_delete = es_delete

    def chunk_log(self, n):
        self.logger.info(u"{}: {}.".format(n, resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1000))

    def load(self, **kwargs):
        """Writes data to the elasticsearch instance

        The data contained in the `processed_data`` attribute of
        the `etl` instance are written into the elasticsearch instance,
        using the specified index and doc_type

        """

        self.logger.setLevel(getattr(logging, self.log_level.upper(), logging.WARNING))

        es = Elasticsearch(self.es_hosts)

        if es.indices.exists(self.es_index):
            if self.es_delete:
                result = es.indices.delete(self.es_index)
                self.logger.info(u'Cancellazione indice "{}": {}.'.format(self.es_index, result))
        else:
            result = es.indices.create(self.es_index)
            self.logger.info(u'Creazione indice "{}": {}.'.format(self.es_index, result))

        # need to wait a couple of seconds before index is available
        # after deletion
        time.sleep(2)

        if self.es_mapping:
            result = es.indices.put_mapping(index=self.es_index, doc_type=self.es_doc_type, body=self.es_mapping)
            self.logger.info(u'Inserimento mapping su indice "{}": {}.'.format(self.es_index, result))

        self.logger.info(u"Import dati ...")
        n = 0

        for n, (ok, result) in enumerate(
            streaming_bulk(
                es, self.parse_data(), index=self.es_index, doc_type=self.es_doc_type, chunk_size=self.es_batchsize
            ),
            1,
        ):
            action, result = result.popitem()
            doc_id = "/{_index}/{_type}/{_id}".format(**result)

            if ok:
                self.logger.debug(
                    u'{}: Operazione "{}" riuscita sul documento {}: ' u"{}.".format(n, action, doc_id, result)
                )
            else:
                self.logger.error(
                    u'{}: Operazione "{}" fallita sul documento {}: ' u"{}.".format(n, action, doc_id, result)
                )

            if n % self.es_batchsize == 0:
                self.chunk_log(n)

        if n % self.es_batchsize != 0:
            self.chunk_log(n)

        self.logger.info(u"Fatto.")

    def parse_data(self):
        for n, row in self.etl.processed_data.iterrows():
            if self.es_id_field in row:
                row["_id"] = row[self.es_id_field]
            yield dict(row)


class DjangoBulkLoader(Loader):
    def __init__(self, django_model, chunk_size=10000, batch_size=1000):
        super().__init__()
        self.django_model = django_model
        self.chunk_size = chunk_size
        self.batch_size = batch_size

    def load(self, **kwargs):
        self.logger.info("Loading data using Django ORM...")

        dictionary = self.etl.processed_data.to_dict("records")

        self.logger.debug(
            "Attempting to import {0} records with bulk_create, "
            "with chunk size of {1} and batch size of {2}.".format(len(dictionary), self.chunk_size, self.batch_size)
        )

        for n, c in enumerate(range(0, len(dictionary), self.chunk_size)):
            model_instances = [self.django_model(**record) for record in dictionary[c: c + self.chunk_size]]
            try:
                self.django_model.objects.bulk_create(model_instances, batch_size=self.batch_size)
                self.logger.info("{0}/{1}".format(n * self.chunk_size, len(dictionary)))
            except Exception as e:
                self.logger.error(e)
                self.logger.error("Loading aborted.")
                return

        self.logger.info("{0}/{0}".format(len(dictionary)))
        self.logger.info(
            "Successfully imported {0} records into {1} table.".format(
                len(dictionary), self.django_model._meta.db_table
            )
        )


class DjangoUpdateOrCreateLoader(Loader):
    def __init__(self, django_model, fields_to_update: Iterable = tuple()):
        super().__init__()
        self.django_model = django_model
        self.fields_to_update = fields_to_update

    def load(self, **kwargs):
        self.logger.info("Loading data using Django ORM...")

        fields_to_update = [f for f in self.fields_to_update if f in self.etl.processed_data.columns]

        records = self.etl.processed_data.to_dict(orient="records")
        self.logger.debug("Attempting to import {0} records.".format(len(records)))
        c = 0
        for record in records:
            defaults = {k: v for k, v in record.items() if k in fields_to_update}
            new_values = {k: v for k, v in record.items() if k not in defaults}
            obj, created = self.django_model.objects.update_or_create(defaults=defaults, **new_values)
            c = c + (1 if created else 0)

        table = self.django_model._meta.db_table
        n_created = c
        n_updated = len(records) - c
        if n_created > 0:
            self.logger.info("Created {0} records into {1} table.".format(n_created, table))
        if n_updated > 0:
            self.logger.info("Updated {0} records into {1} table.".format(n_updated, table))


class LoaderException(Exception):
    pass



# -*- coding: utf-8 -*-
import logging

from .transformations import DummyTransformation

__version__ = '1.1.7'
__description__ = "A framework to write ETL processes at the Openpolis Foundation."

"""A framework to write ETL processes at the Openpolis Foundation.

.. moduleauthor:: Guglielmo Celata <guglielmo@openpolis.it>

Its main scope is to make it easy for the developers to write parsing,
transformation and loading code, re-using logic wherever possible
and writing custom code only when needed.
"""


class ETL(object):
    """Base ETL class

    All ETL instances may be instantiated from this one.

    The class can be instantiated passing an :class:`Extractor`, a :class:`Transformation` and a
    :class:`Loader` instances, as arguments to the constructor::

            etl = ETL(
                extractor=ZIPCSVExctractor(
                        remote_url
                ),
                transformation=RegionCSV2CSVTransformation(),
                loader=CSVLoader(csv_path, label='governi')
            )


    ETL classes support method chaining::

        etl.extract().transform().load()

    or the equivalent shortcut::

        etl.etl()

    The :func:`extract()`, :func:`transform()` and :func:`load()` methods are just wrappers around
    the same methods of the respective
    :class:`Extractor`, :class:`Transformation` and :class:`Loader` instances. These wrappers
    return the `self` instance, in order to allow method chaining.

    """
    logger = logging.getLogger(__name__)

    def __init__(self, extractor, loader, transformation=None,
                 verbosity=False, log_level=logging.INFO, logger=None, source=None, **kwargs):
        self.verbosity = verbosity
        self.source = source

        self.extractor = None
        self.loader = None
        self.transformation = None

        self.original_data = None
        self.processed_data = {}

        if logger:
            self.logger = logger
            self.logger.setLevel(log_level)

        # ooetl ETL subclasses are instantiated without
        # the transformation arguments, so this defaults to
        # the DummyTransformation, in order to be back-compatible
        # the transform method override will still work
        if not transformation:
            transformation = DummyTransformation()

        self.set_extractor(extractor)
        self.set_loader(loader)
        self.set_transformation(transformation)

    def __call__(self):
        """Shortcut for ``extract().transform().load()`` as call operator.

        May be used when there are no parameters to be passed to the
        intermediate methods.
        """
        return self.etl()

    def set_loader(self, loader):
        self.loader = loader
        self.loader.etl = self
        self.loader.logger = self.logger
        self.loader.logger.setLevel(self.logger.getEffectiveLevel())

    def set_transformation(self, transformation):
        self.transformation = transformation
        self.transformation.etl = self
        self.transformation.logger = self.logger
        self.transformation.logger.setLevel(self.logger.getEffectiveLevel())

    def set_extractor(self, extractor):
        self.extractor = extractor
        self.extractor.etl = self
        self.extractor.logger = self.logger
        self.extractor.logger.setLevel(self.logger.getEffectiveLevel())

    def extract(self, **kwargs):
        """Extracts data, using the associated extractor.

        Delegates extraction details to the :func:`Extractor.extract()`
        method of the associated :class:`Extractor` instance.

        Puts extracted data, in the form of :class:`pandas.DataFrame` into
        the `original_data` attribute.

        Returns:
            ETL: self, to allow method chaining.
        """
        self.original_data = self.extractor.extract(**kwargs)
        return self

    def load(self, **kwargs):
        """Loads data, using the associated loader.

        Delegates loading details to the :func:`Loader.load()` method of the
        associated :class:`Loader` instance. Data are fetched from the
        `processed_data` attribute of the class.

        Returns:
            ETL: self, to allow method chaining
        """
        self.loader.load(**kwargs)
        return self

    def transform(self, **kwargs):
        """Transforms the data, using the associated transformation

        Delegates loading details to the :func:`Transformation.transform()` method of the
        associated :class:`Transform` instance.
        Transforms the `original_data` fetched
        by the extractor and fill the `processed_data` attribute, used by the loader.

        `original_data` keeps the original parsed data
        `processed_data` stores the processed data,
        as single `DataFrame` or as a dictionary of multiple
        `DataFrame`, with defined labels.

        Version 1.0 ooetl ETL subclasses, directly overriding the transform method
        are back-compatible

        Returns:
            ETL: self, to allow method chaining
        """
        self.transformation.transform(**kwargs)
        return self

    def etl(self):
        """ Shortcut for ``extract().transform().load()``

        May be used when there are no parameters to be passed to the
        intermediate methods
        """
        return self.extract().transform().load()



# -*- coding: utf-8 -*-
""" Extractors classes

An :class:`Extractor` class should be instantiated
with a remote URL or a local path.

Once instantiated, invoking the :func:`extract()` method will open the resource,
extract the information into a :class:`pandas.DataFrame` and return it to
the caller.

The extract method, when invoked for an instance pointing to a local resource,
will not use the ``request`` module to open the resource, but the ``open``
python call.
"""
import logging

from io import BytesIO as _io

import zipfile
from abc import ABCMeta, abstractmethod

import pandas as pd
import requests


class Extractor(object):
    """Base abstract class. All Extractors classes extend from this.
    """

    __metaclass__ = ABCMeta

    etl = None
    logger = logging.getLogger(__name__)

    @abstractmethod
    def extract(self, **kwargs):
        """Base extract method, must be instantiated in extended
        classes.
        """
        pass


class DataframeExtractor(Extractor):
    """A dummy extractor, that
    extract data from already in-memory dataframe (thus the nam edummy).

    This can be useful to build ETL processes where multiple sources are extracted and
    processed in memory, before the ETL is defined.
    """

    def __init__(self, df):
        """Create a new instance of the extractor, with datafram in memory.
        """
        self.df = df

    def extract(self):
        return self.df


class RemoteExtractor(Extractor):
    """Base :class:`Extractor` for remote data.

    Extractors reading data from a remote source need a
    `remote_url`, to be specified in the constructor.
    """

    remote_url = ""

    __metaclass__ = ABCMeta

    def __init__(self, url):
        """Create a new instance, setting the the `remote_url` attribute to
        the value passed as `url` argument.

        Args:
            url (string): remote location with data

        Returns:
            instance of a :class:`RemoteExtractor` subclass
        """
        super(RemoteExtractor, self).__init__()
        self.remote_url = url

    def extract(self):
        raise Exception("extract method cannot be defined in abstract class")


class CSVExtractor(Extractor):
    """:class:`Extractor`  for remote data, exposed in an
    uncompressed csv format
    """

    def __init__(self, source, **kwargs):
        """Create a new instance of the extractor

        Needs a remote url and may accept other named arguments,
        the arguments used here are a short list of arguments that can be
        found in the :function:`pandas.read_csv()` method.


        Args:
            source: (string)
            sep: (Optional[string]): The separator character; defaults to ";"
            skiprows: (Optional[string]): The number of rows to skip;
              defaults to 0
            header: (Optional[string]): The row at which the header is
              found; defaults to 0
            encoding: (Optional[string]): Encoding used while reading the
              CSV; defaults to "utf8"
            keep_default_na: (Optiona[bool]): If na_values are specified
              and keep_default_na is False the default NaN values are
             overridden, otherwise they’re appended to.
            dtype: (Optional[dict]): The types of objects extracted, as
              a dict, eg: {'a': np.float64, 'b': str, 'c': np.int32},
              defaults to 'str'
            converters : dict, default None
                Dict of functions for converting values in certain columns.
                Keys can either be integers or column labels
        """
        self.source = source
        self.sep = kwargs.get("sep", ";")
        self.skiprows = kwargs.get("skiprows", 0)
        self.skipfooter = kwargs.get("skipfooter", 0)
        self.header = kwargs.get("header", 0)
        self.encoding = kwargs.get("encoding", "utf8")
        self.dtype = kwargs.get("dtype", "str")
        self.converters = kwargs.get("converters", {})
        self.na_filter = kwargs.get("na_filter", True)
        self.na_values = kwargs.get(
            "na_values",
            [
                "",
                "#N/A",
                "#N/A N/A",
                "#NA",
                "-1.#IND",
                "-1.#QNAN",
                "-NaN",
                "-nan",
                "1.#IND",
                "1.#QNAN",
                "N/A",
                "NA",
                "NULL",
                "NaN",
                "nan",
            ],
        )
        self.keep_default_na = kwargs.get("keep_default_na", True)

    @property
    def remote_url(self):
        """Return self.set_url when required, for compatibility

        :return:
        """
        return self.source

    def extract(self, **kwargs):
        """Extracts data from a remote, csv source

        Args:
            **kwargs: Arbitrary keyword arguments.

        Returns:
            :class:`pandas.DataFrame`: The dataframe containing extracted items
        """

        df = pd.read_csv(
            self.source,
            index_col=False,
            sep=self.sep,
            skiprows=self.skiprows,
            skipfooter=self.skipfooter,
            header=self.header,
            encoding=self.encoding,
            dtype=self.dtype,
            converters=self.converters,
            na_filter=self.na_filter,
            na_values=self.na_values,
            keep_default_na=self.keep_default_na,
        )

        return df


class LocalCSVExtractor(CSVExtractor):
    """:class:`Extractor`  for local data, exposed in an
    uncompressed csv format

    DEPRECATED: kept here for compatibility

    Use CSVExtractor, passing the local_path as the url parameter
    """

    def __init__(self, abs_path, **kwargs):
        """Create a new instance of the extractor

        Needs a remote url and may accept other named arguments,
        the arguments used here are a short list of arguments that can be
        found in the :function:`pandas.read_csv()` method.


        Args:
            abs_path: (string)
            sep: (Optional[string]): The separator character; defaults to ";"
            skiprows: (Optional[string]): The number of rows to skip;
              defaults to 0
            header: (Optional[string]): The row at which the header is
              found; defaults to 0
            encoding: (Optional[string]): Encoding used while reading the
              CSV; defaults to "utf8"
            keep_default_na: (Optiona[bool]): If na_values are specified
              and keep_default_na is False the default NaN values are
             overridden, otherwise they’re appended to.
            dtype: (Optional[dict]): The types of objects extracted, as
              a dict, eg: {'a': np.float64, 'b': str, 'c': np.int32},
              defaults to 'str'
            converters : dict, default None
                Dict of functions for converting values in certain columns.
                Keys can either be integers or column labels
        """
        super(LocalCSVExtractor, self).__init__(abs_path, **kwargs)


class ZIPCSVExtractor(CSVExtractor):
    """:class:`Extractor` for remote data, exposed in a compressed csv format
    """

    def __init__(self, source, **kwargs):
        """Constructor method

        :param source: the source URL of the zipped file
        :param kwargs: kwargs used by CSVExtractor and ``verify_tls``,
            used to avoid https errors due to TLS verification failures
        """
        self.verify_tls = kwargs.pop('verify_tls', True)
        super().__init__(source, **kwargs)

    def extract(self, **kwargs):
        """Extracts data from remote, zipped csv source

        Args:
            **kwargs: Arbitrary keyword arguments.

        Returns:
            :class:`pandas.DataFrame`: The dataframe containing extracted items
        """
        r = requests.get(self.source, verify=self.verify_tls)
        z = zipfile.ZipFile(_io(r.content))
        z_filename = [f.filename for f in z.filelist if f.filename[-4:] == '.csv'][0]
        df = pd.read_csv(
            z.open(z_filename),
            index_col=False,
            sep=self.sep,
            skiprows=self.skiprows,
            header=self.header,
            encoding=self.encoding,
            dtype=self.dtype,
            converters=self.converters,
            na_filter=self.na_filter,
            na_values=self.na_values,
            keep_default_na=self.keep_default_na,
        )

        return df


class HTMLParserExtractor(RemoteExtractor):
    """Base :class:`Extractor` class for data contained in a remote URL as HTML
    code.
    """

    def parse(self, html_content):
        """Parses the HTML content.

        Must be overriden by classes extending this one. Details and
        techniques of parsing are completely delegated to extending classes.

        Args:
            html_content (str): The HTML content to be parsed.

        Returns:
            list of dict: A list of dictionaries containing parsed items.

            The list can be passed to the :func:`pandas.DataFrame()` constructor to build a
            :class:`pandas.DataFrame`.
        """
        raise Exception("parse method cannot be launched in abstract class")

    def extract(self, verbosely=False, **kwargs):
        """Extracts html content from the remote URL.

        Delegates parsing of the HTML code to the
        :func:`HTMLParserExtractor.parse()` method.

        Args:
            verbosely (Optional[bool]): Verbosity during extraction.
                Sets the `etl.verbosity` flag to true, and allows the parser
                to use it internally, in order to print to screen or log,
                for debugging purposes
            **kwargs: Arbitrary keyword arguments.

        Returns:
            :class:`pandas.DataFrame`: The dataframe containing parsed items
        """
        r = requests.get(self.remote_url)

        old_verbosity = self.etl.verbosity
        if verbosely:
            self.etl.verbosity = True
        parsed_items = self.parse(r.content)
        self.etl.verbosity = old_verbosity

        return pd.DataFrame(parsed_items)


class SqlExtractor(Extractor):
    def __init__(self, conn_url, sql, index_col=None):
        """Create a new instance, setting the the `conn_url` attribute to
        the value passed as `conn_url` argument.

        Args:
            conn_url (string): connection url to the db (eg:
                               "mysql://root:@localhost:3306/fb")
            sql (string): string SQL query or SQLAlchemy
                          Selectable (select or text object)
                          to be executed, or database table name.
            index_col (string): name of the column to be used as index

        Returns:
            instance of an :class:`Extractor` subclass
        """
        super(SqlExtractor, self).__init__()
        self.conn_url = conn_url
        self.sql = sql
        if index_col:
            self.index_col = index_col

    def extract(self, **kwargs):
        return pd.read_sql(self.sql, self.conn_url, coerce_float=False)


class XLSExtractor(Extractor):
    """:class:`Extractor`  for remote data, exposed in an
    uncompressed csv format
    """

    def __init__(self, io, **kwargs):
        """Create a new instance of the extractor

        Needs a remote url and may accept other named arguments,
        the arguments used here are a short list of arguments that can be
        found in the :function:`pandas.read_excel()` method.


        Args:
            io: (string), file-like object, pandas ExcelFile,
                or xlrd workbook. The string could be a URL.
                Valid URL schemes include http, ftp, s3, and file.
                For file URLs, a host is expected.
                For instance, a local file could be
                file://localhost/path/to/workbook.xlsx
            sheet_name: (Optional[string]): The sheet name, defaults to 0;
                the first sheet.
            header: (Optional[string]): The row at which the header is
              found; defaults to 0
            skiprows: (Optional[string]): The number of rows to skip;
              defaults to 0
            skipfooter: (Optional[string]): The number of rows to skip;
              at the bottom of the file; defaults to 0
            keep_default_na: (Optiona[bool]): If na_values are specified
              and keep_default_na is False the default NaN values are
             overridden, otherwise they’re appended to.
            dtype: (Optional[dict]): The types of objects extracted, as
              a dict, eg: {'a': np.float64, 'b': str, 'c': np.int32},
              defaults to 'str'
            converters : dict, default None
                Dict of functions for converting values in certain columns.
                Keys can either be integers or column labels
        """

        super(XLSExtractor, self).__init__()
        self.io = io
        self.sheet_name = kwargs.get("sheet_name", 0)
        self.skiprows = kwargs.get("skiprows", 0)
        self.skipfooter = kwargs.get("skipfooter", 0)
        self.header = kwargs.get("header", 0)
        self.dtype = kwargs.get("dtype", "str")
        self.converters = kwargs.get("converters", {})
        self.na_filter = kwargs.get("na_filter", True)
        self.na_values = kwargs.get(
            "na_values",
            [
                "",
                "#N/A",
                "#N/A N/A",
                "#NA",
                "-1.#IND",
                "-1.#QNAN",
                "-NaN",
                "-nan",
                "1.#IND",
                "1.#QNAN",
                "N/A",
                "NA",
                "NULL",
                "NaN",
                "nan",
            ],
        )
        self.keep_default_na = kwargs.get("keep_default_na", True)

    def extract(self, **kwargs):
        """Extracts data from a remote, csv source

        Args:
            **kwargs: Arbitrary keyword arguments.

        Returns:
            :class:`pandas.DataFrame`: The dataframe containing extracted items
        """

        df = pd.read_excel(
            io=self.io,
            skiprows=self.skiprows,
            skipfooter=self.skipfooter,
            sheet_name=self.sheet_name,
            header=self.header,
            dtype=self.dtype,
            converters=self.converters,
            na_filter=self.na_filter,
            na_values=self.na_values,
            keep_default_na=self.keep_default_na,
        )

        return df


class FakeExtractor(Extractor):
    """Extractor that does nothing"""

    def extract(self, kwargs) -> pd.DataFrame:
        """Do nothing

        Args:
            **kwargs: Arbitrary keyword arguments.

        Returns:
            :class:`pandas.DataFrame`: The dataframe containing nothing
        """
        return pd.DataFrame()



import logging
from abc import ABCMeta, abstractmethod


class Transformation(object):
    """Base abstract class. All Transformation classes extend from this.
    """

    __metaclass__ = ABCMeta

    etl = None
    logger = logging.getLogger(__name__)

    @abstractmethod
    def transform(self, **kwargs):
        """Base transform method, must be instantiated in extended
        classes.
        """
        pass


class DummyTransformation(Transformation):
    """A dummy transformation that implements a neutral transform method
    original_data are passed to processed_data.

    This is implemented so that ETL subclasses written with ooetl 1.x
    are still compatible with the 2.x releases.
    """
    def transform(self):
        self.etl.processed_data = self.etl.original_data



# -*- coding: utf-8 -*-

try:
    import django.test
except ImportError as error:
    print(f"Warning - {error.__class__.__name__}: {error.name}." )

from unittest.mock import patch


class ETLTest(django.test.TestCase):

    @classmethod
    def setUpClass(cls):
        super(ETLTest, cls).setUpClass()
        cls.mock_get_patcher = patch('ooetl.extractors.requests.get')
        cls.mock_get = cls.mock_get_patcher.start()

    @classmethod
    def tearDownClass(cls):
        cls.mock_get_patcher.stop()
        super(ETLTest, cls).tearDownClass()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ooetl', 'ooetl.docs', 'ooetl.extractors', 'ooetl.tests']

package_data = \
{'': ['*'], 'ooetl.docs': ['_static/images/*', 'src/*']}

install_requires = \
['pandas>=1.3.0,<2.0.0', 'requests>=2.26.0,<3.0.0']

extras_require = \
{':extra == "elastic" or extra == "all"': ['elasticsearch>=7.13.3,<8.0.0'],
 ':extra == "html" or extra == "all"': ['lxml>=4.6.3,<5.0.0',
                                        'cssselect>=1.1.0,<2.0.0'],
 ':extra == "myssql" or extra == "all"': ['mysqlclient>=2.0.3,<3.0.0'],
 ':extra == "myssql" or extra == "postgresql" or extra == "all"': ['SQLAlchemy'],
 ':extra == "postgresql" or extra == "all"': ['psycopg2-binary>=2.9.1,<3.0.0'],
 ':extra == "sparql" or extra == "all"': ['SPARQLWrapper>=1.8.5,<2.0.0']}

setup_kwargs = {
    'name': 'ooetl',
    'version': '1.1.7',
    'description': 'Minimal opinionated object oriented ETL framework',
    'long_description': '[![Latest Version](https://img.shields.io/pypi/v/ooetl.svg)](https://pypi.python.org/pypi/ooetl)\n[![Latest Version](https://img.shields.io/pypi/pyversions/ooetl.svg)](https://pypi.python.org/pypi/ooetl)\n[![License](https://img.shields.io/pypi/l/ooetl.svg)](https://pypi.python.org/pypi/ooetl)\n[![Downloads](https://pepy.tech/badge/ooetl/month)](https://pepy.tech/project/ooetl/month)\n\n[![Twitter Follow](https://img.shields.io/twitter/follow/openpolislab)](https://twitter.com/openpolislab)\n\n![Tests Badge](https://op-badges.s3.eu-west-1.amazonaws.com/ooetl/tests-badge.svg?2)\n![Coverage Badge](https://op-badges.s3.eu-west-1.amazonaws.com/ooetl/coverage-badge.svg?2)\n![Flake8](https://op-badges.s3.eu-west-1.amazonaws.com/ooetl/flake8-badge.svg?2)\n\n\n**ooetl** is a minimal opinionated object oriented ETL framework.\n\nThe class-based nature of the framework allows to build complex dedicated classes,\nstarting from simple abstract ones.\n\n\n## Installation\n\nPython versions from 3.7.1 are supported.\n\nThe package is hosted on pypi, and can be installed, for example using pip:\n\n    pip install --upgrade "ooetl[all]"\n    pip install "ooetl[elastic]==1.1.2"\n\nor poetry:\n\n    poetry add ooetl -Eall\n    poetry add ooetl -Eelastic\n\n\n## Usage\n\nThis is a quick tutorial on how to use the `ooetl` module in order to fetch\ndata from a SQL source and store them into a CSV destination.\n\nThe ETL process is performed by invoking the `etl()` method on a `ooetl.ETL` instance.\n\nThe `etl()` method is a shortcut to the sequence `ooetl.ETL::extract().transform().load()`,\nwhich is possible, as each method returns a pointer to the `ooetl.ETL` instance.\n\nWhen the `ooetl.ETL` instance invokes the `ooetl.ETL::extract()` method, it invokes the corresponging\n`ooetl.Extractor::extract()` method of the *extractor*. The method extracts the data from the source\ninto the `ooetl.ETL::original_data` attribute of the ooetl.`ETL` instance.\n\nThe `ooetl.ETL::transform()` method is overridden in the instance and may be used to apply\ncustom data transformation, before the loading phase.\nThe data from `ooetl.ETL::original_data` are then transformed into `ooetl.ETL::processed_data`.\n\nThe `ooetl.ETL::load()` method invokes the `ooetl.Loader::load()` method storing the data from\n`ooetl.ETL::processed_data` into the defined destination.\n\nThe package provides a series of simple Extractors and Loaders, derived from common abstract classes.\n\nExtractors:\n\n- CSVExctractor - extracts data from a remote CSV\n- ZIPCSVExctractor - extracts data from a remote zipped CSV\n- HTMLParserExtractor - extracts data from a remote HTML page (requires html extra and needs to be extended)\n- SparqlExtractor - extracts data from a remote SPARQL endpoint (requires sparql extra)\n- SqlExtractor - extracts data from a RDBMS (requires mysql or postgresql extra)\n\nLoaders:\n\n- CSVLoader - loads data into a CSV\n- JsonLoader - loads data into a json file\n- ESLoader - loads data into an ES instance (requires elastic extra)\n- DjangoBulkLoader - adds data in bulk to a django model (only works inside a django project)\n- DjangoUpdateOrCreateLoader - adds data with an update or create logic into a django model (slow, only works within a django project)\n\nThe `ooetl.ETL` abstract class is defined in the `__init__.py` file of the `ooetl` package.\n\nETL classes implement a pipeline of extraction, transformation and load logic.\n\nAa a very basic example, here is how to extract data from a postgresql query, into a CSV file.\n\n```python\n    from ooetl import ETL\n    from ooetl.extractors import SqlExtractor\n    from ooetl.loaders import CSVLoader\n    \n    ETL(\n        extractor=SqlExtractor(\n            conn_url="postgresql://postgres:@localhost:5432/opdm",\n            sql="select id, name, inhabitants from popolo_area where istat_classification=\'COM\' "\n                "order by inhabitants desc"\n        ),\n        loader=CSVLoader(\n            csv_path="./",\n            label=\'opdm_areas\'\n        )\n    )()\n```\n\nExtractors (and Loaders) may be easily extended within the projects using the `ooetl` package.\nAs an example, consider the following snippet, extending the abstract `ooetl.HTMLParserExtractor`, that parser\nthe Italian government\'s site and extracts the list of officials, as CSV.\n\nThis example requires the html extra to be installed.\n\n```python\n    import requests\n    from lxml import html\n    from lxml.cssselect import CSSSelector\n    \n    from ooetl import ETL\n    from ooetl.extractors import HTMLParserExtractor\n    from ooetl.loaders import CSVLoader\n    \n    class GovernoExtractor(HTMLParserExtractor):\n    \n        def parse(self, html_content):\n            list_tree = html.fromstring(html_content)\n            items = []\n            for e in CSSSelector(\'div.content div.box_text a\')(list_tree):\n                item_name = e.text_content().strip()\n                item_url = e.get(\'href\').strip()\n                item_page = requests.get(item_url)\n                item_tree = html.fromstring(item_page.content)\n                item_par = CSSSelector(\'div.content div.field\')(item_tree)[0]\n                item_charge = CSSSelector(\'blockquote p\')(item_par)[0].text_content().strip()\n                item_descr = " ".join([\n                  e.text_content() for e in CSSSelector(\'p\')(item_par)[1:] if\\\n                     e.text_content() is not None\n                ])\n                items.append({\n                    \'nome\': item_name,\n                    \'url\': item_url,\n                    \'incarico\': item_charge,\n                    \'descrizione\': item_descr\n                })\n    \n                print(item_name)\n    \n            return items\n    \n    ETL(\n        extractor=GovernoExtractor("https://www.governo.it/it/ministri-e-sottosegretari"),\n        loader=CSVLoader(\n            csv_path="./",\n            label=\'governo\'\n        )\n    )()\n```\n\nOther, more complex examples can be found in the `examples` directory.\n\n## Support\n\nThere is no guaranteed support available, but authors will try to keep up with issues \nand merge proposed solutions into the code base.\n\n## Project Status\nThis project is currently being developed by the [Openpolis Foundation](https://www.openpolis.it/openpolis-foundation/)\nand is being used interanally.\n\nCurrently extras for elasticsearch and sparql have been developed.\n \nShould more be needed, you can either ask to increase the coverage, or try to contribute, following instructions below.\n\n## Contributing\nIn order to contribute to this project:\n* verify that python 3.7.1+ is being used (or use [pyenv](https://github.com/pyenv/pyenv))\n* verify or install [poetry](https://python-poetry.org/), to handle packages and dependencies in a leaner way, \n  with respect to pip and requirements\n* clone the project `git clone git@github.com:openpolis/ooetl.git` \n* install the dependencies in the virtualenv, with `poetry install -Eall`,\n  this will also install the dev dependencies and all extras\n* develop and test \n* create a [pull request](https://docs.github.com/en/github/collaborating-with-pull-requests/proposing-changes-to-your-work-with-pull-requests/about-pull-requests)\n* wait for the maintainers to review and eventually merge your pull request into the main repository\n\n### Testing\nTests are under the tests folder, and can be launched with \n\n    pytest\n\nRequests and responses from ATOKA\'s API are mocked, in order to avoid having to connect to \nthe remote service during tests (slow and needs an API key).\n\nCoverage is installed as a dev dependency and can be used to see how much of the package\'s code is covered by tests:\n\n    coverage run -m pytest\n\n    # sends coverage report to terminal\n    coverage report -m \n\n    # generate and open a web page with interactive coverage report\n    coverage html\n    open htmlcov/index.html \n\nSyntax can be checked with `flake8`.\n\nCoverage and flake8 configurations are in their sections within `setup.cfg`.\n\n## Authors\nGuglielmo Celata - guglielmo@openpolis.it\n\n## Licensing\nThis package is released under an MIT License, see details in the LICENSE.txt file.\n',
    'author': 'guglielmo',
    'author_email': 'guglielmo@openpolis.it',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/openpolis/ooetl/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



import pandas as pd

try:
    from SPARQLWrapper import SPARQLWrapper
except ImportError as error:
    print(f"Warning - {error.__class__.__name__}: {error.name}." )

from ooetl.extractors import RemoteExtractor


class SparqlExtractor(RemoteExtractor):
    def __init__(self, url, query):
        """Create a new instance, setting the the `remote_url` attribute to
        the value passed as `url` argument.

        Args:
            url (string): remote location with data

        Returns:
            instance of a :class:`RemoteExtractor` subclass
        """
        super(SparqlExtractor, self).__init__(url)
        self.remote_url = url
        self.sparql_wrapper = SPARQLWrapper(self.remote_url)
        self.query = query

    def extract(self, **kwargs):
        self.sparql_wrapper.setQuery(self.query)
        self.sparql_wrapper.setReturnFormat("json")
        results = self.sparql_wrapper.query().convert()

        # transform results into a simple dict
        # so that it can be directly used to create a DataFrame
        res = []
        for r in results["results"]["bindings"]:
            res.append(dict((k, r[k]["value"]) for k in r.keys()))
        return pd.DataFrame(res)
